/*
 * Trie in Text
 *
 * *issues
 *  +correct API
 *   -iterator vs callback
 *  +robustness
 *   -error detection
 *   -auto correction
 *   -concurrent access
 *  +efficiency
 *   -lower memory consumption
 *   -disk space?
 */
/*
 * API
 *  trie_open()
 *  trie_close()
 *  trie_add()
 *  trie_find()
 *  trie_delete()
 *  trie_find_next_key()
 *  trie_traverse()
 *  trie_find_prefix()
 *  trie_print_array()
 */
/*
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "tt.h"

/* configs */
#define OBJ_LEN 10
#define LINE_LEN 64
#define CACHE_SIZE 16

/* cell type */
#define TT_SUPER 0
#define TT_UNUSED 1
#define TT_ALLOCED 2
#define TT_NODE 3
#define TT_BODY 4
#define TT_TAIL 5

/* cell structure */
struct cell {
  /* (common) type */
  int type;
  int gc;
  int idx;
  /* union */
  union {
    /* unused */
    int next_unused;
    /* super */
    struct {
      int first_unused;
      int root_cell;
      int size;
    } super;
    /* node */
    struct {
      int key;
      int next;
      int child;
      int body;
      int parent;
    } node;
    /* body */
    struct {
      int owner;
      char *obj;
    } body;
    /* tail */
    struct {
      char *obj;
      int prev;
    } tail;
  } u;
  /* body & tail */
  int next_tail;
};

struct text_trie {
  /**/
  FILE *rfp;
  FILE *wfp;
  /* cache */
  struct cell *super;
  /**/
  struct cell *cache[CACHE_SIZE];
};

static void
print_super_cell(struct cell *c)
{
  printf("super first_unused=%d, root_cell=%d, size=%d\n",
	 c->u.super.first_unused, c->u.super.root_cell, c->u.super.size);
}

static void
print_alloced_cell(struct cell *c)
{
  printf("alloc-ed\n");
}

static void
print_node_cell(struct cell *c)
{
  printf("node key=%d", c->u.node.key);
  if (c->u.node.key > 0 && isprint(c->u.node.key)) {
    printf("(%c)", c->u.node.key);
  }
  printf(" parent=%d next=%d child=%d body=%d\n",
	 c->u.node.parent, c->u.node.next, c->u.node.child, c->u.node.body);
}

static void
print_unused_cell(struct cell *c)
{
  printf("unused next_unused=%d\n",
	 c->u.next_unused);
}

static void
print_body_cell(struct cell *c)
{
  printf("body object=(%s), owner=%d, next_tail=%d\n",
	 c->u.body.obj, c->u.body.owner, c->next_tail);
}

static void
print_tail_cell(struct cell *c)
{
  printf("tail object=(%s), prev=%d, next_tail=%d\n",
	 c->u.tail.obj, c->u.tail.prev, c->next_tail);
}

static void
print_cell(int idx, struct cell *c)
{
  if (!c) {
    printf("idx =%d(null cell):\n", idx);
    return ;
  }
  printf("idx=%d(gc=%d):", idx, c->gc);
  switch (c->type) {
  case TT_SUPER:
    print_super_cell(c);
    break;
  case TT_ALLOCED:
    print_alloced_cell(c);
    break;
  case TT_NODE:
    print_node_cell(c);
    break;
  case TT_UNUSED:
    print_unused_cell(c);
    break;
  case TT_BODY:
    print_body_cell(c);
    break;
  case TT_TAIL:
    print_tail_cell(c);
    break;
  default:
    printf("unknown\n");
  }
}


static char
hex_dec(int x)
{
  const char p[] = "0123456789abcdef";
  return p[x&15];
}

static void
init_buffer(int ln, char *buf)
{
  int i;
  for (i = 0; i < LINE_LEN; i++) {
    buf[i] = ' ';
  }
  buf[0] = hex_dec(ln);
}

static void
trim_buffer(char *buf)
{
  int i;
  for (i = 0; i < LINE_LEN; i++) {
    if (buf[i] == 0 || buf[i] == '\n') {
      buf[i] = ' ';
    }
  }
  buf[LINE_LEN-1] = '\n';
}

static void
encode_super(struct cell *c, char *buf)
{
  buf ++;
  buf += sprintf(buf, "Super ");
  buf += sprintf(buf, "size=%d ", c->u.super.size);
  buf += sprintf(buf, "root=%d ", c->u.super.root_cell);
  buf += sprintf(buf, "unuse=%d ", c->u.super.first_unused);
  buf += sprintf(buf, "w=%d ", LINE_LEN);
}

static void
encode_node(struct cell *c, char *buf)
{
  buf ++;
  buf += sprintf(buf, "Node%c%c ", hex_dec(c->u.node.key>>4),
		 hex_dec(c->u.node.key));
  buf += sprintf(buf, "p=%d nx=%d ch=%d b=%d",
		 c->u.node.parent, c->u.node.next,
		 c->u.node.child, c->u.node.body);
}

static void
encode_body(struct cell *c, char *buf)
{
  buf ++;
  sprintf(buf, "Bnx=%d ow=%d ob=\"%s\"",
	  c->next_tail, c->u.body.owner, c->u.body.obj);
}

static void
encode_unused(struct cell *c, char *buf)
{
  buf++;
  sprintf(buf, "-next=%d", c->u.next_unused);
}

static void
encode_tail(struct cell *c, char *buf)
{
  buf++;
  sprintf(buf, "Tpr=%d nx=%d \"%s\"",
	  c->u.tail.prev, c->next_tail, c->u.tail.obj);
}

static void
encode_unknown(struct cell *c, char *buf)
{
  buf ++;
  sprintf(buf, "?");
}

static void
encode_cell(struct cell *c, char *buf)
{
  switch (c->type) {
  case TT_SUPER:
    encode_super(c, buf);
    break;
  case TT_NODE:
    encode_node(c, buf);
    break;
  case TT_UNUSED:
    encode_unused(c, buf);
    break;
  case TT_BODY:
    encode_body(c, buf);
    break;
  case TT_TAIL:
    encode_tail(c, buf);
    break;
  default:
    encode_unknown(c, buf);
    break;
  }
}

static void
write_back_cell(struct text_trie *a, struct cell *c)
{
  char buf[LINE_LEN+1];
  fseek(a->wfp, c->idx * LINE_LEN, SEEK_SET);
  init_buffer(c->idx, buf);
  encode_cell(c, buf);
  trim_buffer(buf);
  fwrite(buf, LINE_LEN, 1, a->wfp);
  fflush(a->wfp);
}

static void
release_cell(struct text_trie *a, struct cell *c)
{
  if (!c) {
    return ;
  }
  /*free(c);*/
}

/* destructive! */
static char *
decode_str(char *buf)
{
  char *head;
  while (*buf && *buf != '\"') {
    buf ++;
  }
  if (!*buf) {
    /* cant find double quote */
    return strdup("");
  }
  buf ++;
  head = buf;
  while(*buf) {
    buf ++;
  }
  while(*buf != '\"') {
    buf--;
  }
  *buf = 0;
  return strdup(head);
}

static struct cell *
alloc_cell_cache(struct text_trie *a)
{
  struct cell *c = malloc(sizeof(struct cell));
  c->gc = 0;
  return c;
}

static struct cell *
decode_super(struct text_trie *a, char *buf)
{
  struct cell *c = alloc_cell_cache(a);
  c->type = TT_SUPER;
  buf += 7;
  sscanf(buf, "size=%d root=%d unuse=%d",
	 &c->u.super.size, &c->u.super.root_cell, &c->u.super.first_unused);
  return c;
}

static struct cell *
decode_unuse(struct text_trie *a, char *buf)
{
  struct cell *c = alloc_cell_cache(a);
  c->type = TT_UNUSED;
  buf += 2;
  sscanf(buf, "next=%d", &c->u.next_unused);
  return c;
}

static struct cell *
decode_node(struct text_trie *a, char *buf)
{
  struct cell *c = alloc_cell_cache(a);
  c->type = TT_NODE;
  buf+= 5;
  sscanf(buf, "%x p=%d nx=%d ch=%d b=%d",
	 &c->u.node.key, &c->u.node.parent, &c->u.node.next,
	 &c->u.node.child, &c->u.node.body);
  return c;
}

static struct cell *
decode_body(struct text_trie *a, char *buf)
{
  struct cell *c = alloc_cell_cache(a);
  c->type = TT_BODY;
  buf += 2;
  sscanf(buf, "nx=%d ow=%d",
	 &c->next_tail, &c->u.body.owner);
  c->u.body.obj = decode_str(buf);
  return c;
}

static struct cell *
decode_tail(struct text_trie *a, char *buf)
{
  struct cell *c = alloc_cell_cache(a);
  c->type = TT_TAIL;
  buf += 2;
  sscanf(buf, "pr=%d nx=%d",
	 &c->u.tail.prev, &c->next_tail);
  c->u.tail.obj = decode_str(buf);
  return c;
}

static struct cell *
decode_alloced(struct text_trie *a, char *buf)
{
  struct cell *c = alloc_cell_cache(a);
  c->type = TT_ALLOCED;
  return c;
}

static struct cell *
decode_nth_cell_fp(struct text_trie *a, int nth)
{
  struct cell *c;
  char buf[LINE_LEN+1];
  int res;
  fseek(a->rfp, nth * LINE_LEN, SEEK_SET);
  res = fread(buf, LINE_LEN, 1, a->rfp);
  if (res == 0) {
    return NULL;
  }

  buf[LINE_LEN-1] = 0;
  switch (buf[1]) {
  case 'S':
    return decode_super(a, buf);
  case '-':
    return decode_unuse(a, buf);
  case 'N':
    return decode_node(a, buf);
  case 'B':
    return decode_body(a, buf);
  case 'T':
    return decode_tail(a, buf);
  case '?':
    return decode_alloced(a, buf);
  default:
    printf("decode fail (nth=%d::%s).\n", nth, buf);
  }
  c = malloc(sizeof(struct cell));
  c->type = TT_UNUSED;
  return c;
}

static struct cell *
find_cell_from_cache(struct text_trie *tt, int nth)
{
  int i;
  struct cell *c = NULL;
  /* search */
  for (i = 0; i < CACHE_SIZE; i++) {
    if (tt->cache[i] &&
	tt->cache[i]->idx == nth) {
      c = tt->cache[i];
      tt->cache[i] = NULL;
      break;
    }
  }
  if (!c) {
    return NULL;
  }
  /* reorder to make LRU order */
  for (i = CACHE_SIZE - 1; i > 0; i--) {
    if (!tt->cache[i]) {
      tt->cache[i] = tt->cache[i-1];
    }
  }
  tt->cache[0] = c;
  return c;
}

static void
add_cell_to_cache(struct text_trie *tt, struct cell *c)
{
  int i;
  release_cell(tt, tt->cache[CACHE_SIZE-1]);
  for (i = CACHE_SIZE - 1; i > 0; i--) {
    tt->cache[i] = tt->cache[i-1];
  }
  tt->cache[0] = c;
}

/* treat 0 as special to find bug */
static struct cell *
get_nth_cell(struct text_trie *tt, int nth)
{
  struct cell *c;
  if (nth == 0) {
    return NULL;
  }
  c = find_cell_from_cache(tt, nth);
  if (c) {
    return c;
  }
  c = decode_nth_cell_fp(tt, nth);
  if (!c) {
    return NULL;
  }
  c->idx = nth;
  add_cell_to_cache(tt, c);
  return c;
}

static struct cell *
create_super(struct text_trie *a)
{
  struct cell *super;
  super = alloc_cell_cache(a);
  /**/
  super->idx = 0;
  super->type = TT_SUPER;
  super->u.super.first_unused = 0;
  super->u.super.root_cell = 0;
  super->u.super.size = 1;
  return super;
}

static struct cell *
get_super_cell(struct text_trie *tt)
{
  /* cached? */
  if (tt->super) {
    return tt->super;
  }
  /* read */
  tt->super = decode_nth_cell_fp(tt, 0);
  if (tt->super) {
    return tt->super;
  }
  /* create now */
  tt->super = create_super(tt);
  return tt->super;
}

/* convenience function */
static int
get_array_size(struct text_trie *a)
{
  struct cell *super = get_super_cell(a);
  int size = super->u.super.size;
  return size;
}
/* convenience function */
static int
get_root_idx(struct text_trie *a)
{
  struct cell *super = get_super_cell(a);
  int root_idx = super->u.super.root_cell;
  return root_idx;
}

static void
ensure_array(struct text_trie *a, int len)
{
  int i;
  struct cell *super;
  int size = get_array_size(a);
  if (size >= len) {
    return ;
  }
  /* resize the array */
  super = get_super_cell(a);
  for (i = super->u.super.size; i < len; i++) {
    struct cell *ex_cell = alloc_cell_cache(a);
    ex_cell->idx = i;
    ex_cell->type = TT_UNUSED;
    ex_cell->u.next_unused = super->u.super.first_unused;
    ex_cell->gc = 0;
    ex_cell->idx = i;
    write_back_cell(a, ex_cell);
    super->u.super.first_unused = i;
  }
  super->u.super.size = len;
  write_back_cell(a, super);
}

void
trie_print_array(struct text_trie *tt)
{
  int i;
  int size = get_array_size(tt);
  print_cell(0, get_super_cell(tt));
  for (i = 1; i < size; i++) {
    struct cell *c = get_nth_cell(tt, i);
    print_cell(i, c);
  }
}

static int
alloc_cell(struct text_trie *a)
{
  struct cell *super;
  int unuse;
 again:
  super = get_super_cell(a);
  unuse = super->u.super.first_unused;
  if (unuse) {
    struct cell *new_cell = get_nth_cell(a, unuse);
    super->u.super.first_unused = new_cell->u.next_unused;
    new_cell->type = TT_ALLOCED;
    write_back_cell(a, new_cell);
    write_back_cell(a, super);
    return unuse;
  }
  /* expand array */
  ensure_array(a, super->u.super.size + 1);
  goto again;
}

static void
free_cell(struct text_trie *a, int idx)
{
  struct cell *super = get_super_cell(a);
  struct cell *c = get_nth_cell(a, idx);
  c->type = TT_UNUSED;
  c->u.next_unused = super->u.super.first_unused;
  super->u.super.first_unused = idx;
  write_back_cell(a, c);
  write_back_cell(a, super);
}

static struct text_trie *
trie_alloc()
{
  int r;
  struct text_trie *a;
  struct cell *super, *root;
  r = alloc_cell(a);
  /**/
  super = get_super_cell(a);
  super->u.super.root_cell = r;
  write_back_cell(a, super);
  /**/
  root = get_nth_cell(a, r);
  root->type = TT_NODE;
  root->u.node.parent = 0;
  write_back_cell(a, root);
  return a;
}

static void
load_super(struct text_trie *a)
{
  struct cell *root, *super;
  int root_idx;
  super = get_super_cell(a);
  /**/
  if (a->super->u.super.root_cell) {
    return ;
  }
  /**/
  root_idx = alloc_cell(a);
  root = alloc_cell_cache(a);
  root->idx = root_idx;
  root->u.node.key = 0;
  root->type = TT_NODE;
  root->u.node.parent = 0;
  root->u.node.next = 0;
  root->u.node.child = 0;
  root->u.node.body = 0;
  write_back_cell(a, root);
  /**/
  a->super->u.super.root_cell = root_idx;
  write_back_cell(a, super);
}

static void
init_cache(struct text_trie *tt)
{
  int i;
  for (i = 0; i < CACHE_SIZE; i++) {
    tt->cache[i] = NULL;
  }
}

static void
purge_cache(struct text_trie *tt)
{
}

/* API */
struct text_trie *
trie_open(const char *fn)
{
  struct text_trie *a;
  FILE *fp, *wfp;
  /**/
  wfp = fopen(fn, "w");
  if (!wfp) {
    return NULL;
  }
  fp = fopen(fn, "r");
  setvbuf(fp, NULL, _IONBF, 0);
  if (!fp) {
    fclose(wfp);
    return NULL;
  }
  a = malloc(sizeof(struct text_trie));
  a->rfp = fp;
  a->wfp = wfp;
  a->super = NULL;
  /**/
  load_super(a);
  init_cache(a);
  /**/
  return a;
}

/* API */
void
trie_close(struct text_trie *tt)
{
  fclose(tt->rfp);
  fclose(tt->wfp);
}

static void
graft_child(struct text_trie *a, int parent_idx, int new_idx)
{
  struct cell *parent_cell;
  struct cell *new_cell;
  struct cell *cur_youngest_cell;
  int cur_idx;
  /**/
  parent_cell = get_nth_cell(a, parent_idx);
  /**/
  if (parent_cell->u.node.child == 0) {
    /* 1st child */
    parent_cell->u.node.child = new_idx;
    write_back_cell(a, parent_cell);
    return ;
  }

  cur_youngest_cell = get_nth_cell(a, parent_cell->u.node.child);
  new_cell = get_nth_cell(a, new_idx);
  if (new_cell->u.node.key < cur_youngest_cell->u.node.key) {
    /* newly added younger child */
    new_cell->u.node.next = parent_cell->u.node.child;
    parent_cell->u.node.child = new_idx;
    write_back_cell(a, new_cell);
    write_back_cell(a, parent_cell);
    return;
  }

  /* insert some order */
  cur_idx = parent_cell->u.node.child;
  while (cur_idx) {
    struct cell *cur_cell = get_nth_cell(a, cur_idx);
    int next_idx = cur_cell->u.node.next;
    struct cell *next_cell = NULL;
    if (next_idx) {
      next_cell = get_nth_cell(a, next_idx);
    }
    if (!next_cell) {
      /* append */
      new_cell->u.node.next = 0;
      cur_cell->u.node.next = new_idx;
      write_back_cell(a, cur_cell);
      break;
    } else {
      if (cur_cell->u.node.key < new_cell->u.node.key &&
	  new_cell->u.node.key < next_cell->u.node.key) {
	cur_cell->u.node.next = new_idx;
	new_cell->u.node.next = next_idx;
	write_back_cell(a, cur_cell);
	break;
      }
    }
    cur_idx = next_idx;
  }
  write_back_cell(a, new_cell);
}

static int
find_child(struct text_trie *a, int parent_idx, char key)
{
  int child_idx;
  struct cell *parent_cell;
  parent_cell = get_nth_cell(a, parent_idx);
  child_idx = parent_cell->u.node.child;
  while (child_idx) {
    struct cell *child_cell;
    child_cell = get_nth_cell(a, child_idx);
    if (child_cell->u.node.key == key) {
      return child_idx;
    }
    child_idx = child_cell->u.node.next;
  }
  return 0;
}

static int
trie_search_rec(struct text_trie *a, const char *key,
		int parent_idx, int create)
{
  struct cell *child_cell;
  int child_idx;
  /* special case */
  if (!*key) {
    return parent_idx;
  }
  /* scan child */
  child_idx = find_child(a, parent_idx, *key);
  if (!child_idx) {
    if (!create) {
      return 0;
    }
    /* add child */
    child_idx = alloc_cell(a);
    child_cell = get_nth_cell(a, child_idx);
    child_cell->type = TT_NODE;
    child_cell->u.node.parent = parent_idx;
    child_cell->u.node.key = *key;
    child_cell->u.node.next = 0;
    child_cell->u.node.child = 0;
    child_cell->u.node.body = 0;
    write_back_cell(a, child_cell);
    /* graft */
    graft_child(a, parent_idx, child_idx);
  }
  key ++;
  if (!*key) {
    return child_idx;
  }
  return trie_search_rec(a, key, child_idx, create);
}

static char *
get_str_part(const char *str, int from)
{
  char buf[OBJ_LEN+1];
  int i;
  for (i = 0; i < OBJ_LEN; i++) {
    buf[i] = str[from+i];
  }
  buf[i] = 0;
  return strdup(buf);
}

static void
release_body(struct text_trie *a, int idx)
{
  struct cell *c = get_nth_cell(a, idx);
  int tail_idx = c->next_tail;
  while (tail_idx) {
    struct cell *tail_cell = get_nth_cell(a, tail_idx);
    int tmp;
    tmp = tail_cell->next_tail;
    free_cell(a, tail_idx);
    tail_idx = tmp;
  }
  free_cell(a, idx);
}

static void
set_body(struct text_trie *a, int idx, const char *body)
{
  int body_idx = alloc_cell(a);
  int len;
  int i;
  struct cell *node_cell;
  struct cell *body_cell;
  struct cell *prev_cell;
  int prev_idx;
  /**/
  node_cell = get_nth_cell(a, idx);
  if (node_cell->u.node.body) {
    release_body(a, node_cell->u.node.body);
  }
  len = strlen(body);
  /**/
  node_cell->u.node.body = body_idx;
  write_back_cell(a, node_cell);
  /**/
  body_cell = get_nth_cell(a, body_idx);
  body_cell->type = TT_BODY;
  body_cell->u.body.obj = get_str_part(body, 0);
  body_cell->u.body.owner = idx;
  body_cell->next_tail = 0;
  write_back_cell(a, body_cell);
  body_cell = get_nth_cell(a, body_idx);
  /**/
  prev_idx = body_idx;
  prev_cell = body_cell;
  for (i = OBJ_LEN; i < len; i += OBJ_LEN) {
    int tail_idx = alloc_cell(a);
    struct cell *tail_cell = get_nth_cell(a, tail_idx);
    tail_cell->type = TT_TAIL;
    tail_cell->u.tail.obj = get_str_part(body, i);
    tail_cell->u.tail.prev = prev_idx;
    tail_cell->next_tail = 0;
    prev_cell->next_tail = tail_idx;
    write_back_cell(a, tail_cell);
    write_back_cell(a, prev_cell);
    /**/
    prev_idx = tail_idx;
    prev_cell = tail_cell;
  }
  write_back_cell(a, body_cell);
}

/* API */
void
trie_add(struct text_trie *tt, const char *key, const char *body)
{
  int root_idx = get_root_idx(tt);
  int target_idx;
  target_idx = trie_search_rec(tt, key, root_idx, 1);
  if (target_idx) {
    set_body(tt, target_idx, body);
  }
}

static char *
gather_str(struct text_trie *a, int body_idx)
{
  struct cell *body_cell;
  int idx = body_idx;
  char *buf;
  int len = 0;
  /* count length */
  while (idx) {
    struct cell *c = get_nth_cell(a, idx);
    idx = c->next_tail;
    len += OBJ_LEN;
  }
  /**/
  buf = malloc(len + 1);
  idx = body_idx;
  len = 0;
  while (idx) {
    struct cell *c = get_nth_cell(a, idx);
    if (len == 0) {
      sprintf(&buf[len], "%s", c->u.body.obj);
    } else {
      sprintf(&buf[len], "%s", c->u.tail.obj);
    }
    idx = c->next_tail;
    len += OBJ_LEN;
  }
  return buf;
}

/* API */
char *
trie_find(struct text_trie *tt, char *key)
{
  int root_idx = get_root_idx(tt);
  int target_idx;
  target_idx = trie_search_rec(tt, key, root_idx, 0);
  if (target_idx) {
    struct cell *target_cell = get_nth_cell(tt, target_idx);
    int body_idx = target_cell->u.node.body;
    if (body_idx) {
      return gather_str(tt, body_idx);
    }
  }
  return NULL;
}

/* API */
char *
trie_find_next_key(struct text_trie *tt, char *buf, int len)
{
  int root_idx = get_root_idx(tt);
  int target_idx;
  struct cell *target_cell;
  target_idx = trie_search_rec(tt, buf, root_idx, 0);
  /**/
  target_cell = get_nth_cell(tt, target_idx);
  return NULL;
}

static void
trie_traverse_rec(struct text_trie *a,
		  int (*cb)(const char *suffix, const char *str),
		  int idx, char *buf, int buf_len, int cur)
{
  struct cell *cell = get_nth_cell(a, idx);
  int child_idx;

  /* call back */
  if (cell->u.node.body) {
    char *s = gather_str(a, cell->u.node.body);
    if (cur >= buf_len) {
      buf[buf_len - 1] = 0;
    } else {
      buf[cur] = 0;
    }
    if (cb) {
      cb(buf, s);
    } else {
      printf("%s-%s\n", buf, s);
    }
    free(s);
  }

  /* traverse */
  child_idx = cell->u.node.child;
  while (child_idx) {
    cell = get_nth_cell(a, child_idx);
    if (cur < buf_len) {
      buf[cur] = cell->u.node.key;
    }
    trie_traverse_rec(a, cb, child_idx, buf, buf_len, cur + 1);
    child_idx = cell->u.node.next;
  }
}

/* API */
void
trie_traverse(struct text_trie *a, const char *prefix,
	      char *buf, int len,
	      int (*cb)(const char *suffix, const char *str))
{
  int root_idx = get_root_idx(a);
  int target_idx;
  target_idx = trie_search_rec(a, prefix, root_idx, 0);
  if (!target_idx) {
    return ;
  }
  trie_traverse_rec(a, cb, target_idx, buf, len, 0);
}

/* API */
void trie_find_prefix(struct text_trie *tt, const char *str,
		      char *buf, int buf_len,
		      int (*cb)(const char *key, const char *str))
{
  int idx = get_root_idx(tt);
  int i, len = strlen(str);
  for (i = 0; i < len && i < buf_len; i++) {
    struct cell *c;
    idx = find_child(tt, idx, str[i]);
    if (!idx) {
      return ;
    }
    c = get_nth_cell(tt, idx);
    if (!c || c->type != TT_NODE) {
      return ;
    }
    if (c->u.node.body) {
      char *s = gather_str(tt, c->u.node.body);
      if (cb) {
	cb(NULL, s);
      }
      free(s);
    }
  }
}

static void
disconnect(struct text_trie *a, int parent_idx, int target_idx)
{
  struct cell *parent_cell = get_nth_cell(a, parent_idx);
  struct cell *target_cell = get_nth_cell(a, target_idx);
  if (parent_cell->u.node.child == target_idx) {
    /* 1st child */
    parent_cell->u.node.child = target_cell->u.node.next;
    write_back_cell(a, parent_cell);
  } else {
    /* not 1st child */
    int child_idx = parent_cell->u.node.child;
    while (child_idx) {
      struct cell *cur = get_nth_cell(a, child_idx);
      if (cur->u.node.next == target_idx) {
	cur->u.node.next = target_cell->u.node.next;
	write_back_cell(a, cur);
	return ;
      }
    }
  }
}

static void
release_node_rec(struct text_trie *a, int target_idx)
{
  struct cell *target_cell = get_nth_cell(a, target_idx);
  int child_idx;
  if (target_cell->type != TT_NODE) {
    return ;
  }
  if (target_cell->u.node.body) {
    release_body(a, target_cell->u.node.body);
  }
  child_idx = target_cell->u.node.child;
  while (child_idx) {
    struct cell *child_cell = get_nth_cell(a, child_idx);
    int tmp = child_cell->u.node.next;
    release_node_rec(a, child_idx);
    child_idx = tmp;
  }
  free_cell(a, target_idx);
}

/* API */
void
trie_delete(struct text_trie *a, const char *prefix)
{
  struct cell *target_cell;
  int root_idx = get_root_idx(a);
  int target_idx, parent_idx;
  target_idx = trie_search_rec(a, prefix, root_idx, 0);
  if (!target_idx) {
    return ;
  }
  target_cell = get_nth_cell(a, target_idx);
  parent_idx = target_cell->u.node.parent;
  disconnect(a, parent_idx, target_idx);
  /**/
  release_node_rec(a, target_idx);
}
